#!/bin/bash

declare -a gimpReturnValues
declare -a gimpCmdArr
declare -a gimpCmdArrRetReq
declare -A gimpDisplays
declare gLastMake
state=$gimpTmpDir/gimpState
displays=$gimpTmpDir/gimpDisplays
gimpSharpenMacro=70
#gimpLoadSpecimenDir is in use, don't reset

. $HOME/bin/locale.sh

getMakeOfImg(){
   local file="$1"
   local make

   eval $(exiftool -m -p 'make="$Make"'  "$file")
   echo "$make"
}

isCanonImg(){
    local file="$1"
    local geometry


    eval $(exiftool -m -p 'gLastMake="$Make" geometry="$ImageSize"'  "$file")
    if [ "$gLastMake" == "Canon" ]; then
	return 0
    elif [ "$gLastMake" == "ImagingSource" ]; then
	return 1
    elif [ "$gLastMake" == "tisUcview" ]; then
	return 1
    else
	if [ "$geometry" == "1600x1200" ]; then
    	    return 1
	else
    	    return 0
	fi
    fi
}

gimpResponse(){
    local nBytes str
    local arr=($(od -t u1 -N 4))

    if [ "${arr[1]}" != "71" ]; then
	echo "ERROR: first char of gimp response must be \"G\"" >&2
	return -1
    elif [ "${arr[2]}" -ne 0 ]; then
	echo "WARN: gimp error: ${arr[2]}" >&2
    fi
    nBytes=$((${arr[3]}*256+${arr[4]}))

    read -rN $nBytes str
    echo "$str" 
    return
}

gimpRequest(){
    local cmd="$1"
    local len=${#cmd}
    if [ $len -ge 65536 ]; then
	echo "ERROR: script too long for gimp socket" >&2
	return 1
    fi
    local hi=$(printf "%02x" $(($len / 256)))
    local lo=$(printf "%02x" $(($len % 256)))
    echo -ne "G\x${hi}\x${lo}$cmd"
}

gimpCmds(){
    local i=0
    local cmd

    if [ $# -gt 0 ]; then
	gimpCmdArr=("$@")
    fi

    unset gimpReturnValues
    nCmds=${#gimpCmdArr[*]}
    while [ $i -lt $nCmds ]; do
	if [ $i -ne 0 ]; then
	    res=$(gimpResponse)
	    echo "${res}:" >&2
	    gimpReturnValues[$i]="$(cat $gimpTmpDir/gimp.tmp)"
	    echo "${gimpReturnValues[$i]}" >&2
	    echo >&2
	fi
	cmd=$(printf "(let ((fd (open-output-file \"%s\"))) (write %s fd)(close-output-port fd))"  $gimpTmpDir/gimp.tmp "${gimpCmdArr[$i]}")
#echo "$cmd" >&2
	gimpRequest "$cmd"
	i=$(($i+1))
	if [ $i -eq $nCmds ]; then
	    res=$(gimpResponse)
	    echo "${res}:" >&2
	    gimpReturnValues[$i]="$(cat $gimpTmpDir/gimp.tmp)"
	    echo "${gimpReturnValues[$i]}" >&2
	    echo >&2
	fi
    done </dev/tcp/localhost/10008 >&0
    rm $gimpTmpDir/gimp.tmp 2>/dev/null
}

gimpCmdsg(){
    local i
    local cmd

    if [ $# -gt 0 ]; then
	i=0
	for val in "$@"; do
	    if [ $(($i % 2)) -eq 0 ]; then
		gimpCmdArr[$(($i/2))]="$val"
	    else
		if [ "$val" != "y" -a "$val" != "n" ]; then
		    echo "ERROR: every second val in arglist must be \"y\" or \"n\"" >&2
		    return 1
		fi
		gimpCmdArrRetReq[$((($i-1)/2))]="$val"
	    fi
	    i=$(($i+1))
	done
    fi

    i=0
    unset gimpReturnValues
    nCmds=${#gimpCmdArr[*]}
    while [ $i -lt $nCmds ]; do
	if [ $i -ne 0 ]; then
	    res=$(gimpResponse)
	    echo "${res}:" >&2
	    if [ ${gimpCmdArrRetReq[$(($i-1))]} == "y" ]; then
		    gimpReturnValues[$i]="$(cat $gimpTmpDir/gimp.tmp)"
		    echo "${gimpReturnValues[$i]}" >&2
	    fi
	    echo >&2
	fi
#echo "${i}: ${gimpCmdArr[$i]} ${gimpCmdArrRetReq[$i]}" >&2
	if [ ${gimpCmdArrRetReq[$i]} == "y" ]; then
	    cmd=$(printf "(define globalVar %s)(let ((fd (open-output-file \"%s\"))) (write globalVar fd)(close-output-port fd))" "${gimpCmdArr[$i]}"  $gimpTmpDir/gimp.tmp)
	else
	    cmd="${gimpCmdArr[$i]}"
	fi
#echo "$cmd" >&2
	gimpRequest "$cmd"
	i=$(($i+1))
	if [ $i -eq $nCmds ]; then
	    res=$(gimpResponse)
	    echo "${res}:" >&2
	    if [ ${gimpCmdArrRetReq[$(($i-1))]} == "y" ]; then
		gimpReturnValues[$i]="$(cat $gimpTmpDir/gimp.tmp)"
		echo "${gimpReturnValues[$i]}" >&2
	    fi
	    echo >&2
	fi
    done </dev/tcp/localhost/10008 >&0
    rm $gimpTmpDir/gimp.tmp 2>/dev/null
}

gimpCmdListOpenFiles(){
    printf "(let* ((imgs (gimp-image-list)) (nImgs (car imgs)) (imgVec (cadr imgs))) (do ((i 0 (+ i 1))) ((= i nImgs) nImgs) (display (car (gimp-image-get-filename (vector-ref imgVec i))) fd) (newline fd)))"
}

gimpCmdLoadFile(){
    local file="$1"
    printf "(let* ((image (car (gimp-file-load 1 \"$file\" \"/$file\")))) (car (gimp-display-new image)))" "$file"
}

gimpCmdGetImgNo(){
    local file="$1"
    printf "(let* ((imgs (gimp-image-list)) (nImgs (car imgs)) (imgVec (cadr imgs))) (do ((i 0 (+ i 1))) ((or (= i nImgs) (string=? \"%s\" (car (gimp-image-get-filename (vector-ref imgVec i))))) (if (= i nImgs) #f (vector-ref imgVec i))) ()))" "$file"
}

gimpCmdGetDrawable(){
    local file="$1"
    printf "(let* ((img %s)) (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))))" "$(gimpCmdGetImgNo $file)"
}

gimpCmdSetFilename(){
    local oldFile="$1"
    local newFile="$2"
    printf "(let* ((img %s)) (gimp-image-set-filename img \"%s\"))" "$(gimpCmdGetImgNo $oldFile)" "$newFile"
}

gimpCmdSharpen(){
    local percent="$1"
    local file="$2"

    printf "(let* ((img %s)) (plug-in-sharpen 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) %s))" "$(gimpCmdGetImgNo $file)" $percent
}

gimpCmdBlur(){
    local file="$1"

    printf "(let* ((img %s)) (plug-in-gauss 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) 2.0 2.0 1))" "$(gimpCmdGetImgNo $file)" 
}

gimpCmdSaveJpeg(){
    local quality=$1
    local file="$2"
    local targetFile="${3:-$2}"
    
    printf "(let ((img %s)) (gimp-image-merge-visible-layers img 1) (file-jpeg-save 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) \"%s\" \"%s\" $quality 0 1 1 \"\" 1 1 0 1))" "$(gimpCmdGetImgNo $file)" $targetFile $targetFile
}

gimpCmdSaveJpegAndClose(){
    local quality=$1
    local file="$2"
    local targetFile="${3:-$2}"
    
    printf "(let ((img %s)) (gimp-image-merge-visible-layers img 1) (file-jpeg-save 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) \"%s\" \"%s\" $quality 0 1 1 \"\" 1 1 0 1) (gimp-image-delete img))" "$(gimpCmdGetImgNo $file)" $targetFile $targetFile
}

gimpCmdSaveJpegAndDisplay(){
    local quality=$1
    local file="$2"
    local targetFile="${3:-$2}"
    
    printf "(let ((img %s)) (gimp-image-merge-visible-layers img 1) (file-jpeg-save 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) \"%s\" \"%s\" $quality 0 1 1 \"\" 1 1 0 1) (car (gimp-display-new img)))" "$(gimpCmdGetImgNo $file)" $targetFile $targetFile
}

gimpCmdCloseFile(){
    file="$1"
    printf "(gimp-display-delete %s)" ${gimpDisplays[$file]}
}

gimpCmdAddScalebar(){
    local file="$1"
    local width scale length

    eval $(exiftool -m -p 'scale="$ImageDescription" width="$ImageWidth"'  "$file")
    if [ -z "$scale" ]; then
	echo "#f"
	return
    fi

    unit=3
    length=$(calc "10*$scale")
    while [ $(calc "$length > 0.6*$width && $unit > 1") == "1" ]; do
	unit=$(($unit-1))
	length=$(calc "$length/10")
    done
    printf "(let ((img %s)) $(cat ~/bin/insert-scalebar.gimp) (insert-scalebar img %s %s %s))" "$(gimpCmdGetImgNo $file)" $length 0 0 
}

gimpCmdScale(){
    local targetWidth="$1"
    local targetHeight="$2"
    local file="$3"

    printf "(let* ((img %s) (w (car (gimp-image-width img ))) (h (car (gimp-image-height img)))) (if (or (> w $targetWidth) (> h $targetHeight)) (let ((factor (if (>= (/ w h) 1.5) (/ $targetWidth w) (/ $targetHeight h)))) (gimp-image-scale img (* w factor) (* h factor)))))" "$(gimpCmdGetImgNo $file)"
}

gimpCmdScaleWithScalebar(){
    local targetWidth="$1"
    local targetHeight="$2"
    local file="$3"

    printf "(let* ((img %s) (w (car (gimp-image-width img ))) (h (car (gimp-image-height img))) (scalebar (car (gimp-image-get-layer-by-name img \"scalebar\"))) (layers (gimp-image-get-layers img))) (if (or (> w $targetWidth) (> h $targetHeight)) (let ((factor (if (>= (/ w h) 1.5) (/ $targetWidth w) (/ $targetHeight h)))) (if (and (not (= scalebar -1)) (gimp-item-is-layer scalebar)) (let* ((offsets (gimp-drawable-offsets scalebar)) (swidth (car (gimp-drawable-width scalebar))) (color (car (gimp-image-pick-color img scalebar (car offsets) (cadr offsets) FALSE FALSE 1))) (fg (car (gimp-context-get-foreground)))) $(cat ~/bin/insert-scalebar.gimp) (gimp-image-remove-layer img scalebar) (gimp-image-scale img (* w factor) (* h factor)) (gimp-context-set-foreground color) (insert-scalebar img (* factor swidth) (* factor (car offsets)) (* factor (cadr offsets))) (gimp-context-set-foreground fg))  (gimp-image-scale img (* w factor) (* h factor))) )))"  "$(gimpCmdGetImgNo $file)"
}

gimpCmdDuplicate(){
    local originFile="$1"
    local targetFile="$2"
    
    printf "(let* ((origin %s) (target (car (gimp-image-duplicate origin)))) (gimp-image-set-filename target \"%s\") target)" "$(gimpCmdGetImgNo $originFile)" "$targetFile"
}

gimpCmdUnsharpMask(){
    local file="$1"
    local radius=${2:-5.0}
    local amount=${3:-0.3}

    printf "(let* ((img %s) (decomposed (car (plug-in-decompose 1 img (let ((layers (gimp-image-get-layers img))) (vector-ref (cadr layers) (- (car layers) 1))) \"HSV\" 1)))) (plug-in-unsharp-mask 1 decomposed (car (gimp-image-get-active-drawable decomposed)) $radius $amount 0) (plug-in-recompose 1 decomposed (car (gimp-image-get-active-drawable decomposed))) (gimp-image-delete decomposed) img)" "$(gimpCmdGetImgNo $file)"
}

gimpCmdCloseDisplays(){
    local display

    printf "(let ((displays (list %s))) (do ((display displays (cdr display))) ((null? display)) (if (= 1 (car (gimp-display-is-valid (car display)))) (gimp-display-delete (car display)))))" "$*"
}

gimpCmdLevels(){
    local file="$1"
    printf "(let ((drawable (car (gimp-image-get-active-drawable %s)))) (gimp-levels drawable 1 ${gimpLevelsRed}) (gimp-levels drawable 2 ${gimpLevelsGreen}) (gimp-levels drawable 3 ${gimpLevelsBlue}))" "$(gimpCmdGetImgNo $file)"
    
}

gimpCmdLevelsOrig(){
    local file="$1"
    printf "(let ((drawable (car (gimp-image-get-active-drawable %s)))) (gimp-levels drawable 1 0 113 1 0 255) (gimp-levels drawable 2 0 211 1 0 255) (gimp-levels drawable 3 0 164 1 0 255))" "$(gimpCmdGetImgNo $file)"
    
}

gimpCmdLevelsStretch(){
    local file="$1"
    printf "(gimp-levels-stretch (car (gimp-image-get-active-drawable %s)))" "$(gimpCmdGetImgNo $file)"
}

gimpDuplicate(){
    local originFile="$1"
    local targetFile="$2"

    gimpCmds "$(gimpCmdDuplicate $originFile $targetFile)"
}

gimpUnsharpMask(){
    gimpCmds "$(gimpCmdUnsharpMask $1 $2 $3)"
}

gimpScaleFile(){
    local targetWidth="$1"
    local targetHeight="$2"
    local file="$3"

    gimpCmds "$(gimpCmdScale $targetWidth $targetHeight $file)"
}

gimpLevels(){
    local file="$1"

    gimpCmds "$(gimpCmdLevels $file)" "(gimp-displays-flush)"
}

gimpLoadFiles(){
    local file i

    unset gimpCmdArr
    i=0
    for file in "$@"; do
	gimpCmdArr[$i]="$(gimpCmdLoadFile $file)"
	i=$(($i+1))
    done
    gimpCmds

    i=0
    for file in "$@"; do
	i=$(($i+1))
	gimpDisplays[$file]="${gimpReturnValues[$i]}"
    done
}

gimpCloseFiles(){
    local file i

    i=0
    unset gimpCmdArr
    for file in "$@"; do
	if [ -z "${gimpDisplays[$file]}" ]; then
	    echo "WARN: no display found for \"$file\", already closed" >&2
	    continue
	fi
	gimpCmdArr[$i]="$(gimpCmdCloseFile $file)"
	unset gimpDisplays[$file]
	i=$(($i+1))
    done
    if [ ${#gimpCmdArr[*]} -gt 0 ]; then
	gimpCmds
    fi
}

gimpSharpenFiles(){
    local file i

    i=0
    unset gimpCmdArr
    for file in "$@"; do
	gimpCmdArr[$i]="$(gimpCmdSharpen $gimpSharpenMacro $file)"
	i=$(($i+1))
    done
    if [ ${#gimpCmdArr[*]} -gt 0 ]; then
	gimpCmdArr[$i]="(gimp-displays-flush)"
	gimpCmds
    fi
    
}

gimpSaveJpegs(){
    local file i

    i=0
    unset gimpCmdArr
    for file in "$@"; do
	gimpCmdArr[$i]="$(gimpCmdSaveJpeg 0.9 $file)"
	i=$(($i+1))
    done
    if [ ${#gimpCmdArr[*]} -gt 0 ]; then
	gimpCmds
    fi
}

gimpSaveAndCloseFiles(){
    local file i

    i=0
    unset gimpCmdArr
    for file in "$@"; do
	if [ -z "${gimpDisplays[$file]}" ]; then
	    echo "WARN: no display found for \"$file\", already closed" >&2
	    continue
	fi
	gimpCmdArr[$i]="$(gimpCmdSaveJpeg 0.9 $file)"
	i=$(($i+1))
	gimpCmdArr[$i]="$(gimpCmdCloseFile $file)"
	i=$(($i+1))
	unset gimpDisplays[$file]
    done
    if [ ${#gimpCmdArr[*]} -gt 0 ]; then
	gimpCmds
    fi
}

gimpListOpenFiles(){
    gimpCmds "$(gimpCmdListOpenFiles)"
}

gimpWebbifyFiles() {
    local show="n"
    if [ "$1" == "-show" ]; then
	show="y"
	shift
    fi
    local file i
    local webFile webThumb

    i=0
    unset gimpCmdArr
    for file in "$@"; do
	base=$(basename $file)
	webFile="${file#$origDir}"
	if [ "$webFile" == "$file" ]; then
	    #echo "$webFile <-> $file" >&2
	    echo "ERROR: you're probably in the wrong directory" >&2
	    read
	    return 1
	fi
	webFile=${targetDir}$webFile
	webThumb=$(dirname $webFile)/th_$(basename $webFile)

	gimpCmdArr[$i]="$(gimpCmdDuplicate $file $webFile)"
	i=$(($i+1))
	if isCanonImg "$file"; then
	    gimpCmdArr[$i]="#f"
	    i=$(($i+1))
	    gimpCmdArr[$i]="$(gimpCmdScaleWithScalebar 1200 800 $webFile)"
	    i=$(($i+1))
	    gimpCmdArr[$i]="$(gimpCmdDuplicate $webFile $webThumb)"
	    i=$(($i+1))
	    if [ -L "${file}.dontSharpen" ]; then
		gimpCmdArr[$i]="#f"
		i=$(($i+1))
	    else
		gimpCmdArr[$i]="$(gimpCmdSharpen $gimpSharpenMacro $webFile)"
		i=$(($i+1))
	    fi
	    gimpCmdArr[$i]="$(gimpCmdScaleWithScalebar 300 200 $webThumb)"
	    i=$(($i+1))
	    if [ -L "${file}.dontSharpen" ]; then
		gimpCmdArr[$i]="#f"
	    else
		gimpCmdArr[$i]="$(gimpCmdSharpen 20 $webThumb)"
	    fi
	    i=$(($i+1))
	else
	    if [ "$gLastMake" == "ImagingSource" ]; then
		gimpCmdArr[$i]="$(gimpCmdBlur $webFile)"
	    else
		gimpCmdArr[$i]="#f"
	    fi
	    i=$(($i+1))

	    gimpCmdArr[$i]="$(gimpCmdScaleWithScalebar 1200 800 $webFile)"
	    i=$(($i+1))

	    if [ -L "${file}.dontSharpen" ]; then
		gimpCmdArr[$i]="#f"
		i=$(($i+1))
	    else
		gimpCmdArr[$i]="$(gimpCmdUnsharpMask $webFile 5 0.3)"
		i=$(($i+1))
	    fi

	    gimpCmdArr[$i]="$(gimpCmdDuplicate $webFile $webThumb)"
	    i=$(($i+1))
	    gimpCmdArr[$i]="$(gimpCmdScaleWithScalebar 300 200 $webThumb)"
	    i=$(($i+1))
	    if [ -L "${file}.dontSharpen" ]; then
		gimpCmdArr[$i]="#f"
	    else
		gimpCmdArr[$i]="$(gimpCmdUnsharpMask $webThumb 20 1.0)"
	    fi
	    i=$(($i+1))

	fi
	if [ "$show" == "y" ]; then
	    gimpCmdArr[$i]="$(gimpCmdSaveJpegAndDisplay 0.8 $webFile)"
	    i=$(($i+1))
	    gimpCmdArr[$i]="$(gimpCmdSaveJpegAndDisplay 0.8 $webThumb)"
	    i=$(($i+1))
	else
	    gimpCmdArr[$i]="$(gimpCmdSaveJpegAndClose 0.8 $webFile)"
	    i=$(($i+1))
	    gimpCmdArr[$i]="$(gimpCmdSaveJpegAndClose 0.8 $webThumb)"
	    i=$(($i+1))
	fi

    done

    gimpCmds

    if [ "$show" == "y" ]; then
	i=0
	for file in "$@"; do
	    webFile="${file#$origDir}"
	    webFile=${targetDir}$webFile
	    webThumb=$(dirname $webFile)/th_$(basename $webFile)
	    
	    gimpDisplays[$webFile]=${gimpReturnValues[$(($i*9+8))]}
	    gimpDisplays[$webThumb]=${gimpReturnValues[$(($i*9+9))]}
	done
    fi

}

gimpLoadSpecimen(){
    local file i n make
    local dir="$1"
    if ! [ -d "$dir" ]; then
	echo "ERROR: \"$dir\": no such directory" >&2
	return 1
    fi
    if [ "$(echo $dir/*.JPG)" == "$dir/*.JPG" ]; then
	echo "ERROR: \"$dir\": no .JPG files" >&2
	return 1
    fi
    gimpLoadFiles $dir/*.JPG
    gimpLoadSpecimenDir="$dir"
    echo >&2

    i=1
    while [ $i -le ${#gimpReturnValues[*]} ]; do
	echo ${gimpReturnValues[$i]}
	i=$(($i+1))
    done > $displays
    
    unset gimpCmdArr
    i=0
    for file in $dir/*.JPG; do
	if ! [ -L "${file}.noWhiteBalance" ]; then
	    make=$(getMakeOfImg "$file")
	    base=$(basename $file)
	    if [ "$make" == "ImagingSource" ]; then
		gimpCmdArr[$i]="$(gimpCmdLevelsPrev $file)"
		i=$(($i+1))
	    elif [ "$make" == "tisUcview" ]; then
		gimpCmdArr[$i]="$(gimpCmdLevels $file)"
		i=$(($i+1))
	    elif [ "$make" != "Canon" ]; then
		echo "WARN: unknown Make \"$make\"" >&2
	    fi
	fi
	gimpCmdArr[$i]="$(gimpCmdAddScalebar $file)"
	i=$(($i+1))
    done
    if [ $i -gt 0 ]; then
	gimpCmdArr[$i]="(gimp-displays-flush)"
	i=$(($i+1))
	gimpCmds
    fi
}

gimpWebbifySpecimen(){
    gimpListOpenFiles
    local -a openFiles=(${gimpReturnValues[1]})
    unset openFiles[$((${#openFiles[*]}-1))]
    local -A openArr
    local webDir orgFile webThumb i base
    local dir="${1:-$gimpLoadSpecimenDir}"

    if [ -z "$dir" ]; then
	echo "ERROR: no dir" >&2
	return 1
    fi

    i=0
    while [ $i -lt ${#openFiles[*]} ]; do
	openArr[${openFiles[$i]}]="open"
	i=$(($i+1))
    done

    webDir="${targetDir}${dir#$origDir}"
    if [ ! -d "$webDir" ]; then
	mkdir -p $webDir
    else
	for file in $webDir/*.JPG; do
	    if [ "$file" == "$webDir/*.JPG" ]; then
		break
	    fi
	    base="$(basename $file)"
	    if [ "${base#th_}" != "$base" ]; then
		continue
	    fi
	    orgFile=${origDir}${file#$targetDir}
	    if [ "${openArr[$orgFile]}" != "open" ]; then
		unset openArr[$orgFile]
		webThumb=$(dirname $file)/th_$(basename $file)
		if [ -f $webThumb ]; then
		    rm $file $webThumb
		else
		    rm $file
		fi
	    fi
	done
    fi
    
    gimpWebbifyFiles ${openFiles[*]}
}

gimpCloseAll(){
    gimpListOpenFiles
    local -a openFiles=(${gimpReturnValues[1]})
    unset openFiles[$((${#openFiles[*]}-1))]
    local i

    unset gimpCmdArr
    i=0
    while [ $i -lt ${#openFiles[*]} ]; do
	if [ -z "${gimpDisplays[${openFiles[$i]}]}" ]; then
	    echo "ERROR: lost display of file \"${openFiles[$i]}\"" >&2
	else
	    gimpCmdArr[$i]="(gimp-display-delete ${gimpDisplays[${openFiles[$i]}]})"
	fi
	i=$(($i+1))
    
    done
    gimpCmds
    unset gimpDisplays
}

calc () {
    local expr="$1"
    bc <<EOF
scale=12
$expr
quit
EOF
}
